// XML Note Tools
// Jeffrey J Weimer
// SVN History: $Revision: 566 $ on $Date: 2008-03-03 18:15:21 -0500 (Mon, 03 Mar 2008) $

// Accessible Functions

// XMLPrint(property,text,[attributes, values, pretty])
// prints XML string to history
//	property: string of propert(ies)	- "property1:property2: ... propertyN"
//	text: string text to put at final property
//	attributes: string list of attributes		- "attribute1;attribute2; ... attributeN"
//	values: list of values for each attribute
//	pretty: variable 	- 0 (default) returns without tabs and returns, 1 returns with tabs and returns

// XMLPrintf(property,text,[attributes, values, pretty])
// same as above but returns string rather than prints to history

// XMLWaveNote(twave,property,text,[attributes, values, pretty, replace, at])
// same as above but attaches text to wave twave (full path or current folder only!)
//	replace: variable		- 0 (default), N => replace forward at N, -N => replace reverse at N
//	at:	string	"end" (default), "start"
//	for a note containing T total properties, replace=(-N) is the same as replace=(T-N+1)

// XMLNoteText(ntbk,property,text,[attributes, values, pretty, replace, at])
// same as above but puts text to notebook ntbk
//	at:	string	"end" (default), "start", "current"
//	selection always remains immediately after insertion for next addition!

// NT_LoadXMLTemplate(template,[filename])
// loads an XML file as a template (empty places contain #)

// NT_FillXMLTemplate(template,paramlist)
// returns text that is the template filled with paramlist in each #

// VARIATIONS
// text or values prefaced with % are interpreted as STRING commands to be executed
// to override this, use \% (escape code sequence)

#pragma rtGlobals=1
#pragma IgorVersion=6.03
#pragma version = 0.20

Function XMLPrint(property,text,[attributes, values, pretty])
	string property, text, attributes, values
	variable pretty

	string theText
	
	if (ParamIsDefault(pretty))
		pretty = 0
	endif
	
	if (ParamIsDefault(attributes))
		theText = xmlprintf(property,text, pretty=pretty)
	else
		if (ParamIsDefault(values))
			theText = xmlprintf(property, text, attributes=attributes, values="", pretty=pretty)
		else
			theText = xmlprintf(property, text, attributes=attributes, values=values, pretty=pretty)
		endif
	endif
	print theText
	return 0
end

Function/S XMLPrintF(property,text,[attributes, values, pretty])
	string property, text, attributes, values
	variable pretty
	
	string theText
	
	if (ParamIsDefault(pretty))
		pretty = 0
	endif
	
	if (ParamIsDefault(attributes))
		theText = parseXMLinput(property,text, pretty=pretty)
	else
		if (ParamIsDefault(values))
			theText = parseXMLinput(property, text, attributes=attributes, values="", pretty=pretty)
		else
			theText = parseXMLinput(property, text, attributes=attributes, values=values, pretty=pretty)
		endif
	endif
	return theText
end

Function XMLWaveNote(twave,property,text,[attributes, values, pretty, replace, at])
	wave twave
	string property, text, attributes, values, at
	variable pretty, replace

	string theText="", theNote=""
	
	if (!WaveExists(twave))
		return -1
	endif

	if (ParamIsDefault(replace))
		replace = 0
	endif
	
	if (ParamIsDefault(at))
		at = "end"
	endif
	
	switch(replace)
		case 0:
			if (ParamIsDefault(pretty))
				pretty = 0
			endif
			if (ParamIsDefault(attributes))
				theText = xmlprintf(property,text, pretty=pretty)
			else
				if (ParamIsDefault(values))
					theText = xmlprintf(property, text, attributes=attributes, values="", pretty=pretty)
				else
					theText = xmlprintf(property, text, attributes=attributes, values=values, pretty=pretty)
				endif
			endif
			break
		default:
			theNote = note(twave)
			theText = ReplaceXMLProperty(property,theNote,text,index=replace)
			note/k twave
			at = "end"	
			break
	endswitch
	strswitch(at)
		case "start":
			theText = theText + note(twave)
			note/k twave
			break
		default:
			break
	endswitch
	note twave, theText
	return 0
end

Function XMLNoteText(ntbk,property,text,[attributes, values, pretty, replace, at])
	string ntbk, property, text, attributes, values, at
	variable pretty, replace

	string theText
	variable st, en
	
	if (strlen(WinList(ntbk,";","WIN:16"))==0)
		return -1
	endif

	if (ParamIsDefault(replace))
		replace = 0
	endif
	
	if (ParamIsDefault(at))
		at = "end"
	endif
	
	switch(replace)
		case 0:
			if (ParamIsDefault(pretty))
				pretty = 0
			endif
			if (ParamIsDefault(attributes))
				theText = xmlprintf(property,text, pretty=pretty)
			else
				if (ParamIsDefault(values))
					theText = xmlprintf(property, text, attributes=attributes, values="", pretty=pretty)
				else
					theText = xmlprintf(property, text, attributes=attributes, values=values, pretty=pretty)
				endif
			endif
			strswitch(at)
				case "current":
					break
				case "start":
					Notebook $ntbk, selection={startOfFile,startOfFile}
					break
				default:
					Notebook $ntbk, selection={endOfFile,endOfFile}
					break
			endswitch
			notebook $ntbk text=theText
			break
		default:
			ReplaceXMLNtbkProperty(property,ntbk,text,index=replace)
			break
	endswitch
	return 0
end

Function NT_LoadXMLTemplate(template,[filename])
	string template, filename
	
	string cdf = GetDataFolder(1), theText
	variable refNum
	
	SetDataFolder root:
	
	if (!DataFolderExists("Packages:XMLNoteTools"))
		NewDataFolder/O/S Packages
		NewDataFolder/O/S XMLNoteTools
	else
		SetDataFolder Packages:XMLNoteTools
	endif

	if (ParamIsDefault(filename))
		theText="Select an XML file for template " + template
		open/D/R/T=".xml"/M=theText refNum
		if (strlen(S_filename)==0)
			close refNum
			SetDataFolder $cdf
			return -1
		endif
		filename = S_filename
	endif
	
	open/R refNum as filename
	
	string/G $template=""
	
	SVAR tstr = $template
	FSetPos refNum, 0
	do
		FReadLine refNum, theText
		if (strlen(theText)==0)
			break
		endif
		tstr += theText
	while(1)
	close refNum
	SetDataFolder $cdf
	return 0
end

Function/S NT_FillXMLTemplate(template,paramlist)
	string template, paramlist
	
	string rtStr, cdf = GetDataFolder(1), theText, theT
	variable ic=0, nt
	
	theText = "root:Packages:XMLNoteTools:" + template
	
	if (exists(theText)==0)
		NT_LoadXMLTemplate(template)
	endif
	
	SVAR tstr = $theText

	SetDataFolder root:
	
	rtStr = tstr
	
	nt = ItemsInList(paramlist)
	do
		theText=StringFromList(ic,paramlist)
		if (strlen(theText)==0)
			break
		endif
		theText = parseXMLvalue(theText)
		rtStr = ReplaceString("#",rtStr,theText,0,1)
		ic += 1
	while(ic<nt)
	SetDataFolder $cdf
	return rtStr
end

Static Function/S ReplaceXMLProperty(theproperty,theText,theReplace, [start, index])
	string theproperty, theText, theReplace
	variable start, index
	
	variable ic, st, en, jc
	string theStr, rtnStr, reStr, thePStr
	
	if (!ParamIsDefault(index))
		if (abs(index)>ItemsInXML(theproperty,theText))
			return theText
		endif
		if (index>0)
			if (ParamIsDefault(start))
				st = 0
			else
				st = start
			endif
			en = st
			for (ic=0;ic<index;ic+=1)
				en = GetXMLOffset(theproperty,theText,start=en)+1
				if (en < 0)
					return theText
				endif
			endfor
			st = en
		else
			if (ParamIsDefault(start))
				st = INF
			else
				st = start
			endif
			en = st
			thePStr = StringFromList(0,theproperty,":")
			for (ic=0;ic>index;ic-=1)
				en = GetXMLOffset(theproperty,theText,start=en,option=1)-1
				if (en < 0)
					return theText
				endif
				en = GetXMLOffset(thePStr,theText,start=en,option=1)-1
			endfor
			en = GetXMLOffset(theproperty,theText,start=en+1)+1
			st = en
		endif
	else
		st = GetXMLOffset(theproperty,theText,start=st)
	endif
	
	jc = ItemsInList(theproperty,":") - 1
	st = strsearch(theText,">",st)+1
	sprintf thePStr, "</%s>", StringFromList(jc,theproperty,":")
	en = strsearch(theText,thePStr,st)
	sprintf theStr, "%s%s",theText[st,en-1], thePStr
	sprintf reStr, "%s%s", theReplace, thePStr
	rtnStr = ReplaceString(theStr,theText,reStr,0,1)
	return rtnStr
end

Static Function ReplaceXMLNtbkProperty(theproperty,theNtbk,theReplace, [index])
	string theproperty, theNtbk, theReplace
	variable index
	
	variable ic, en
	string thePStr
	
	if (!ParamIsDefault(index))
		if (index>0)
			Notebook $theNtbk, selection={startOfFile,startOfFile}
			for (ic=0;ic<index;ic+=1)
				en = SetXMLNtbkOffset(theproperty,theNtbk)
				if (en < 0)
					return -1
				endif
			endfor
		else
			Notebook $theNtbk, selection={endOfFile,endOfFile}
			thePStr = StringFromList(0,theproperty,":")
			for (ic=0;ic>index;ic-=1)
				en = SetXMLNtbkOffset(thePStr,theNtbk,flag=4)
				GetSelection notebook, $theNtbk, 1
				Notebook $theNtbk, selection={(V_startParagraph,0),(V_startParagraph-1,0)}
				if (en < 0)
					return 0
				endif
			endfor
			thePStr=StringFromList(ItemsInList(theproperty,":")-1,theproperty,":")
			en = SetXMLNtbkOffset(thePStr,theNtbk)
		endif
	else
		Notebook $theNtbk, selection={startOfFile,startOfFile}
		en = SetXMLNtbkOffset(theproperty,theNtbk)
	endif
	Notebook $theNtbk, text=theReplace
end

Static Function GetXMLOffset(theproperty,theText, [start, option])
	string theproperty, theText
	variable start, option
	
	variable ic, nt, st
	string theStr
	
	if (ParamIsDefault(option))
		option=0
	endif
	
	if ((option==1)||(option==3))
		if (ParamIsDefault(start))
			start = INF
		endif
	else
		if (ParamIsDefault(start))
			start = 0
		endif
	endif
	
	st = start
	nt = ItemsInList(theproperty, ":")
	if (nt > 1)
		for (ic=0;ic<nt-1;ic+=1)
			sprintf theStr, "<%s", StringFromList(ic,theproperty,":")
			if (!option)
				st = strsearch(theText,theStr,st)
			else
				st = strsearch(theText,theStr,st,option)
			endif
			if (st < 0)
				return st
			endif
		endfor
		sprintf theStr, "<%s", StringFromList(ic,theproperty,":")
		st = strsearch(theText,theStr,st)
	else
		sprintf theStr, "<%s", StringFromList(0,theproperty,":")
		if (!option)
			st = strsearch(theText,theStr,st)
		else
			st = strsearch(theText,theStr,st,option)
		endif
	endif
	return st
end

Static Function SetXMLNtbkOffset(theproperty,theNtbk, [flag])
	string theproperty, theNtbk
	variable flag
	
	variable ic, nt, st, en, pst, pen
	string theStr, theProp
	
	if (ParamIsDefault(flag))
		flag=-1
	endif
	
	nt = ItemsInList(theproperty, ":")
	if (nt > 1)
		for (ic=0;ic<nt;ic+=1)
			theProp = StringFromList(ic,theproperty,":")
			sprintf theStr, "<%s", theProp
			if (flag<0)
				Notebook $theNtbk, findText={theStr,4}
			else
				Notebook $theNtbk, findText={theStr,2^flag}
			endif
			if (!V_flag)
				return -1
			endif
		endfor
	else
		theProp = StringFromList(0,theproperty,":")
		sprintf theStr, "<%s", theProp
		if (flag<0)
			Notebook $theNtbk, findText={theStr,4}
		else
			Notebook $theNtbk, findText={theStr,2^flag}
		endif
		if (!V_flag)
			return -1
		endif
	endif
	sprintf theStr, "/%s>", theProp
	Notebook $theNtbk, findText={theStr,4}
	GetSelection notebook, $theNtbk, 1
	en = V_startPos-1; pen=V_startParagraph
	sprintf theStr, "<%s",theProp
	Notebook $theNtbk, findText={theStr,2^4}
	GetSelection notebook, $theNtbk, 1
	Notebook $theNtbk, findText={">",2}                            
	GetSelection notebook, $theNtbk, 1
	st = V_endPos; pst = V_endParagraph
	Notebook $theNtbk, selection={(pst,st),(pen,en)}
	return (st)
end

Static Function ItemsInXML(property,theText)
	string property, theText
	
	variable ic=0, en=0
	
	do
		en = GetXMLOffset(property,theText, start=en)+1
		if (en == 0)
			break
		endif
		ic +=1
	while(1)	
	return ic
end

Static Function/S parseXMLinput(property,text,[attributes, values, pretty])
	string property, text, attributes, values
	variable pretty

	string tS0 = "<", ts1 = "</", tE = ">", tR = "\r"
	string tT = "\t", tQ = "\"", tB="/"
	
	string theProp ="", rtnStr = "", theEnd = "", theAtt, itsV
	
	variable ic, nt, nv
	
	if (pretty!=1)
		tR = ""
		tT = ""
	endif
	
	if (ParamIsDefault(attributes))
		theAtt = ""
	else
		nt = ItemsInList(attributes)
		if (nt==1)
			itsV = parseXMLvalue(StringFromList(0,values))
			sprintf theAtt, " %s=%s%s%s",attributes,tQ,itsV,tQ
		else
			nv = ItemsInList(values)
			if (nv < nt)
				values = ""
				for(ic=0;ic<nt;ic+=1)
					values += ";"
				endfor
			endif
			theAtt = ""
			for(ic=0;ic<nt;ic+=1)
				theProp = StringFromList(ic,attributes)
				itsV = parseXMLvalue(StringFromList(ic,values))
				sprintf theAtt, "%s %s=%s%s%s", theAtt, theProp,tQ,itsV,tQ
			endfor
		endif
	endif
	
	nt = ItemsInList(property,":")
	
	switch(nt)
		case 0:
			rtnStr = text
			break
		case 1:
			theProp = StringFromList(0,property,":")
			if (strlen(text)==0)
				sprintf rtnStr, "%s%s%s %s%s", ts0,theProp,theAtt,tB,tE
			else
				sprintf rtnStr, "%s%s%s%s%s%s%s%s", ts0,theProp,theAtt,tE,parseXMLvalue(text),ts1,theProp,tE
			endif
			break
		default:
			for (ic=0;ic<nt-1;ic+=1)
				theProp = StringFromList(ic,property,":")
				theEnd += tT
				sprintf rtnStr, "%s%s%s%s%s%s",rtnStr,ts0,theProp,tE,tR,theEnd
			endfor
			theProp = StringFromList(nt-1,property,":")
			if (strlen(text)==0)
				sprintf rtnStr, "%s%s%s%s %s%s", rtnStr,ts0,theProp,theAtt,tB,tE
			else
				sprintf rtnStr, "%s%s%s%s%s%s%s%s%s", rtnStr,ts0,theProp,theAtt,tE,parseXMLvalue(text),ts1,theProp,tE
			endif
			for (ic=nt-2;ic>=0;ic-=1)
				theEnd = RemoveEnding(theEnd,"\t")
				theProp = StringFromList(ic,property,":")
				sprintf rtnStr, "%s%s%s%s%s%s", rtnStr, tR,theEnd,ts1,theProp,tE
			endfor
			break
	endswitch
	
	return rtnStr
end

Static Function/S parseXMLvalue(vStr)
	string vStr
	
	string rtnStr, theCmd
	
	theCmd = vStr[0,1]
	if (cmpstr(theCmd,"\%")==0)
		rtnStr = vStr[1,strlen(vStr)-1]
		return rtnStr
	endif
	
	if (cmpstr(vStr[0],"%") == 0)
		string/G tmpXMLToolStr
		theCmd =  "tmpXMLToolStr = " + vStr[1,strlen(vStr)-1]
		Execute/Q theCmd
		rtnStr = tmpXMLToolStr
		killstrings/Z tmpXMLToolStr
	else
		rtnStr = vStr
	endif
	
	return rtnStr
end
